import csv
import httpx
import tiktoken
import lxml.html
import lxml.etree
from lxml.etree import ParserError
from ssl import SSLCertVerificationError, SSLError
from databeakers.pipeline import Pipeline, EdgeType, ErrorType
from databeakers.beakers import TempBeaker
from databeakers.http import HttpRequest, HttpResponse
from scrapeghost.preprocessors import CleanHTML
from scrapeghost.errors import TooManyTokens
from openai.error import InvalidRequestError
from scrapeghost.errors import BadStop
from .models import Agency, URL, Int, IdOnly, ScrapeghostResponse
from .ghost import ghost


class CSVSource:
    def __init__(self, filename, datatype):
        self.filename = filename
        self.datatype = datatype

    def __call__(self):
        with open(self.filename) as inf:
            reader = csv.DictReader(inf)
            for line in reader:
                yield self.datatype(**line)


def tiktoken_count(response):
    if response.status_code != 200:
        raise ValueError("response status code is not 200")

    html = response.response_body

    # clean the html
    cleaner = CleanHTML()
    encoding = tiktoken.get_encoding("cl100k_base")
    doc = lxml.html.fromstring(html)
    (doc,) = cleaner(doc)  # returns a 1-item list
    html_again = lxml.html.tostring(doc, encoding="unicode")
    tokens = len(encoding.encode(html_again))

    return Int(int=tokens)


# current thinking, beakers exist within a recipe
recipe = Pipeline("foiaghost", "foiaghost.db")
recipe.add_beaker("agency", Agency)
recipe.add_seed(
    "agencies",
    "agency",
    CSVSource("agencies.csv", Agency),
)
recipe.add_beaker("good_urls", URL)
recipe.add_transform(
    "agency",
    "good_urls",
    lambda a: a.url.startswith("http"),
    edge_type=EdgeType.conditional,
)
recipe.add_beaker("responses", HttpResponse)
recipe.add_transform(
    "good_urls",
    "responses",
    HttpRequest(),
    error_map={
        (
            httpx.HTTPError,
            SSLCertVerificationError,
            SSLError,
        ): "bad_requests"
    },
)


class ProcessRecord:
    def __init__(self, func, params_map):
        self.func = func
        self.params_map = params_map

    def __call__(self, record):
        kwargs = {}
        for param, (beaker_name, field_name) in self.params_map.items():
            kwargs[param] = getattr(record[beaker_name], field_name)
        return self.func(**kwargs)

    def __repr__(self):
        return f"ProcessRecord({self.func.__name__}, {self.params_map})"


def scrapeghost_response(record) -> ScrapeghostResponse:
    sg = ghost.scrape(url_or_html=record["responses"].response_body)
    return ScrapeghostResponse(
        total_cost=sg.total_cost, api_time=sg.api_time, data=sg.data
    )


recipe.add_transform(
    "responses",
    "scrapeghost_response",
    scrapeghost_response,
    whole_record=True,
    error_map={
        (TooManyTokens,): "scrapeghost_too_many_tokens",
        (BadStop,): "scrapeghost_bad_stop",
        (InvalidRequestError, ValueError, ParserError): "scrapeghost_invalid_request",
    },
)
